/*
                OpenSceneGraph Public License, Version 0.0
                ==========================================
                
  Copyright (C) 2002 Robert Osfield.
  
  Everyone is permitted to copy and distribute verbatim copies
  of this licence document, but changing it is not allowed.

                       OPENSCENEGRAPH PUBLIC LICENCE
     TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION

  This library is free software; you can redistribute it and/or modify it
  under the terms of the OpenSceneGraph Public License (OSGPL) version 0.0
  or later.
  
  Notes: the OSGPL is based on the LGPL, with the 4 exceptions laid in in
  the wxWindows section below.  The LGPL in the final section of this
  license.


-------------------------------------------------------------------------------

                wxWindows Library Licence, Version 3
                ====================================

  Copyright (C) 1998 Julian Smart, Robert Roebling [, ...]

  Everyone is permitted to copy and distribute verbatim copies
  of this licence document, but changing it is not allowed.

                       WXWINDOWS LIBRARY LICENCE
     TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION

  This library is free software; you can redistribute it and/or modify it
  under the terms of the GNU Library General Public Licence as published by
  the Free Software Foundation; either version 2 of the Licence, or (at
  your option) any later version.

  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library
  General Public Licence for more details.

  You should have received a copy of the GNU Library General Public Licence
  along with this software, usually in a file named COPYING.LIB.  If not,
  write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA 02111-1307 USA.

  EXCEPTION NOTICE

  1. As a special exception, the copyright holders of this library give
  permission for additional uses of the text contained in this release of
  the library as licenced under the wxWindows Library Licence, applying
  either version 3 of the Licence, or (at your option) any later version of
  the Licence as published by the copyright holders of version 3 of the
  Licence document.

  2. The exception is that you may use, copy, link, modify and distribute
  under the user's own terms, binary object code versions of works based
  on the Library.

  3. If you copy code from files distributed under the terms of the GNU
  General Public Licence or the GNU Library General Public Licence into a
  copy of this library, as this licence permits, the exception does not
  apply to the code that you add in this way.  To avoid misleading anyone as
  to the status of such modified files, you must delete this exception
  notice from such code and/or adjust the licensing conditions notice
  accordingly.

  4. If you write modifications of your own for this library, it is your
  choice whether to permit this exception to apply to your modifications.
  If you do not wish that, you must delete the exception notice from such
  code and/or adjust the licensing conditions notice accordingly.
*/ 
  
//------------------------------------------------------------------------------

#include "osgExportImport.h"

#include <sys/types.h>

#include <maya/MItDependencyNodes.h>
#include <maya/MFnAttribute.h>

#include <maya/MStatus.h>
#include <maya/MPxCommand.h>
#include <maya/MString.h>
#include <maya/MStringArray.h>
#include <maya/MArgList.h>
#include <maya/MGlobal.h>
#include <maya/MSelectionList.h>
#include <maya/MItSelectionList.h>
#include <maya/MPoint.h>
#include <maya/MPointArray.h>
#include <maya/MDagPathArray.h>
#include <maya/MFnPlugin.h>
#include <maya/MFnSet.h>
#include <maya/MFnTransform.h>
#include <maya/MItMeshPolygon.h>
#include <maya/MItMeshVertex.h>
#include <maya/MItMeshEdge.h>
#include <maya/MFloatPoint.h>
#include <maya/MFloatPointArray.h>
#include <maya/MFloatVector.h>
#include <maya/MFloatVectorArray.h>
#include <maya/MFloatArray.h>
#include <maya/MObject.h>
#include <maya/MObjectArray.h>
#include <maya/MPlug.h>
#include <maya/MFnDagNode.h>
#include <maya/MFnDependencyNode.h>
#include <maya/MItDag.h>
#include <maya/MItDependencyGraph.h>
#include <maya/MDistance.h>
#include <maya/MIntArray.h>
#include <maya/MFnLambertShader.h>
#include <maya/MIOStream.h>

//------------------------------------------------------------------------------

#include <osgDB/ReadFile>
#include <osgDB/WriteFile>
#include <osgDB/ImageOptions>
#include <osgDB/FileNameUtils>

#include <osg/Node>
#include <osg/Array>
#include <osg/PrimitiveSet>
#include <osg/Vec2>

#include <osg/StateSet>
#include <osg/Texture>
#include <osg/Texture2D>
#include <osg/Texture3D>
#include <osg/GL>
#include <osg/DisplaySettings>
#include <osg/TexEnv>

#include <osgDB/Registry>
#include <osgDB/ReadFile>

#include <osgUtil/TangentSpaceGenerator>

#include <map>
#include <algorithm>
#include <sstream>

using namespace std;

#if 0
const MString LIGHTMAP_UVSET_NAME = "LightMap";
const MString AUX_DATA_ME_UVSET_NAME = "AuxDataME";
const string  OCCLUDER_PREFIX = "occluder_";
#endif

std::ofstream log_file;

//------------------------------------------------------------------------------
inline void assign(MFloatPoint &point, const osg::Vec3 &vec3)
{
    point[0] = vec3[0];
    point[1] = vec3[1];
    point[2] = vec3[2];
}
//------------------------------------------------------------------------------
inline void assign(osg::Vec3 &vec3, const MPoint &point)
{
    vec3[0] = point[0];
    vec3[1] = point[1];
    vec3[2] = point[2];
}
//------------------------------------------------------------------------------
inline void assign(MVector &vector, const osg::Vec3 &vec3)
{
    vector[0] = vec3[0];
    vector[1] = vec3[1];
    vector[2] = vec3[2];
}
//------------------------------------------------------------------------------
inline void assign(osg::Vec3 &vec3, const MVector &vector)
{
    vec3[0] = vector[0];
    vec3[1] = vector[1];
    vec3[2] = vector[2];
}


/* int2str ------------------------------------------------------------------ */
string int2string( const int& i ) {
  ostringstream stream; stream << i;
  return stream.str();
}

//------------------------------------------------------------------------------

const char *const dummyOptionScript = "";

const char *const osgOptionScript = "osgExportOptions";
const char *const osgDefaultOptions =
"save_as_ive=0;"
"duplicate_coords=1;"
"texture_repeat_mode=1;"
"generate_tangent_space=1;"
"use_binormal_vector=0;"
"use_short_if_possible=0;"
;

//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
////////////////// Maya's functions for supporting plugins //////////////////
//---------------------------------------------------------------------------
MStatus initializePlugin( MObject obj )
{  MFnPlugin plugin( obj, "Alias", "3.0", "Any");

MStatus stat;
// Register the translator with the system
stat =	plugin.registerFileTranslator( "OSG", "none",
                                      OSGTranslator::creator,
                                      (char *)osgOptionScript,
                                      (char *)osgDefaultOptions );  
if (!stat) stat.perror("FAILED to load plugin");

return stat;
}
//---------------------------------------------------------------------------
MStatus uninitializePlugin( MObject obj )
{
    MFnPlugin plugin( obj );
    return plugin.deregisterFileTranslator( "OSG" );
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//////////////// parent 'MPxFileTranslator' inherited members: //////////////
//---------------------------------------------------------------------------
void* OSGTranslator::creator()
{
    return new OSGTranslator();
}

#if 0
//---------------------------------------------------------------------------
MStatus OSGTranslator::reader ( const MFileObject& file,
                               const MString& options,
                               FileAccessMode mode)
{
    MStatus status;
    MString mname = file.fullName(), unitName;
    const char *fname = mname.asChar();

    log_file.open("Maya_osg_plugin.log", ios::out|ios::trunc);
    OSGRead(fname);
    log_file.close();

    status.perror("SAS!");
    return MS::kSuccess;
}
#endif

//---------------------------------------------------------------------------
MStatus OSGTranslator::writer ( const MFileObject& file,
                               const MString& options,
                               FileAccessMode mode )

{
    MStatus status;
    MString mname = file.fullName(), unitName;
    const char *fname = mname.asChar();

    log_file.open("Maya_osg_plugin.log", ios::out|ios::trunc);
    getOptionsValues(options);

    cout << "-----" << endl
         << " Beginning export for " << fname << endl;
    
    OSGWrite(fname, mode);
    log_file.close();

    cout << " Done!" << endl
         << "-------" << endl;
    return MS::kSuccess;
}
//---------------------------------------------------------------------------
bool OSGTranslator::haveReadMethod () const
{
//    return true;
  return false;    
}
//---------------------------------------------------------------------------
bool OSGTranslator::haveWriteMethod () const
{
  return true;
}
//---------------------------------------------------------------------------
MString OSGTranslator::defaultExtension () const
{
  return "osg";
}
//---------------------------------------------------------------------------
MString	OSGTranslator::filter () const
{
  return "*.osg";
}
//---------------------------------------------------------------------------
MPxFileTranslator::MFileKind OSGTranslator::identifyFile (
    const MFileObject& fileName,
    const char* buffer,
    short size) const
{
    const char * name = fileName.name().asChar();
    int   nameLength = strlen(name);

    if ((nameLength > 4) &&
        ( !strcasecmp(name+nameLength-4, ".osg")))
    {
        cerr << "Error: The file " << name << " kCouldBeMyFileType" << endl;
        return kCouldBeMyFileType;
    }
    else
    {
        cerr << "Error: The file " << name << " kNotMyFileType" << endl;
        return kNotMyFileType;
    }
}
//---------------------------------------------------------------------------

//***************************************************************************
//*********************** OSG EXPORT IMPORT MEMBERS *************************
//***************************************************************************
//---------------------------------------------------------------------------
void OSGTranslator::getOptionsValues(const MString &options_string)
{
    // Options parsing

    param_texture_repeat_mode = false; // texture is not in the TILE mode
    param_generate_tangent_space = true;
    param_use_binormal_vector = false;
    param_duplicate_coords = false;
    param_use_short_if_possible = false;

    if (options_string.length() > 0)
    {
        int i, length;
        // Start parsing.
        MStringArray optionList;
        MStringArray theOption;
        options_string.split(';', optionList); // break out all the options.

        length = optionList.length();
        for ( i = 0; i < length; ++i )
        {
            theOption.clear();
            optionList[i].split( '=', theOption );

            if ( theOption[0] == MString("texture_repeat_mode")
                 && theOption.length() > 1 )
            {
                if ( theOption[1].asInt() > 0 )
                    param_texture_repeat_mode = true;
                else
                    param_texture_repeat_mode = false;
            }
            if ( theOption[0] == MString("generate_tangent_space")
                 && theOption.length() > 1 )
            {
                if ( theOption[1].asInt() > 0 )
                    param_generate_tangent_space = true;
                else
                    param_generate_tangent_space = false;
            }
            if ( theOption[0] == MString("use_binormal_vector")
                 && theOption.length() > 1 )
            {
                if ( theOption[1].asInt() > 0 )
                    param_use_binormal_vector = true;
                else
                    param_use_binormal_vector = false;
            }
            if ( theOption[0] == MString("duplicate_coords")
                 && theOption.length() > 1 )
            {
                if( theOption[1].asInt() > 0 )
                    param_duplicate_coords = true;
                else
                    param_duplicate_coords = false;
            }
            if ( theOption[0] == MString("use_short_if_possible")
                 && theOption.length() > 1 )
            {
                if ( theOption[1].asInt() > 0 )
                    param_use_short_if_possible = true;
                else
                    param_use_short_if_possible = false;
            }
        }
    }
}
//---------------------------------------------------------------------------
MObject OSGTranslator::findShader( MObject& setNode )
//  Description: 
//      Find the shading node for the given shading group set node.
{
    MFnDependencyNode fnNode(setNode);
    MPlug shaderPlug = fnNode.findPlug("surfaceShader");

    if ( !shaderPlug.isNull() )
    {			
        MPlugArray connectedPlugs;
        bool asSrc = false;
        bool asDst = true;
        shaderPlug.connectedTo( connectedPlugs, asDst, asSrc );

        if ( connectedPlugs.length() != 1 )
            log_file << "Error getting shader" << endl;
        else 
            return connectedPlugs[0].node();
    }			

    return MObject::kNullObj;
}
//---------------------------------------------------------------------------
void OSGTranslator::outNonTriangulatedNodesError()
{
    string error_msg =
      "ERROR: This scene contains these non-triangulated nodes:\n\n";
    log_file << "\n\n" << error_msg;

    for (unsigned int i = 0; i < non_triangulated_nodes.size(); ++i)
    {
        error_msg += non_triangulated_nodes[i];
        error_msg += "\n";

        log_file << non_triangulated_nodes[i] << endl;
    }

    log_file << endl;

    log_file << error_msg << " : Error on exporting to OSG format!" ;
}
//---------------------------------------------------------------------------
void OSGTranslator::OSGWrite(const char *fname, FileAccessMode mode)
{
    string full_file_name = fname;

    osg::ref_ptr<osg::Node> scene_node;

    if ( getMayaSceneAsOSGNode( scene_node, mode ) && is_scene_correct )
    {
        log_file << "\nwriting file: " << full_file_name << endl;
        osgDB::writeNodeFile( *scene_node.get(), full_file_name );
        log_file << "export has been finished!" << endl;
    }
    else
    {
        outNonTriangulatedNodesError();
        log_file << "export aborted!" << endl;
    }
}	


//---------------------------------------------------------------------------
bool OSGTranslator::getMayaSceneAsOSGNode(osg::ref_ptr<osg::Node> &scene_node,
                                          FileAccessMode mode)
{
    log_file << "getMayaSceneAsOSGNode" << endl;
    using namespace osg;

    osg::ref_ptr<osg::Geode> scene_geode;
    osg::ref_ptr<osg::Geode> occluders_geode;

    scene_geode			= new osg::Geode();
    scene_node			= scene_geode.get();

    scene_node->setName("Exported From Maya");

    non_triangulated_nodes.clear();
    is_scene_correct = true;

    MSelectionList list,list1;

    MStatus stat;
    MItDag dagIter( MItDag::kBreadthFirst, MFn::kInvalid, &stat );

    MDagPath dagPath;
    MObject Obj;

    if (mode == kExportActiveAccessMode) 
    {
        MGlobal::getActiveSelectionList (list);	

        for (; !dagIter.isDone(); dagIter.next())
        {
            if (dagIter.getPath (dagPath) != MS::kSuccess)
              continue;

            MFnDagNode dagNode( dagPath, &stat );

            if(dagPath.hasFn(MFn::kObjectNameFilter))
              continue;
            
            // We want only the shape, not the transform-extended-to-shape (OBJ)
            if ( dagPath.hasFn(MFn::kMesh) && dagPath.hasFn(MFn::kTransform) )
              continue;

            if (!dagPath.hasFn( MFn::kMesh ) )
              continue;

            MFnMesh fnMesh(dagPath);

            for(unsigned int i=0;i<list.length();i++)
            {
                list.getDependNode(i,Obj);
                if(fnMesh.hasParent(Obj)) list1.add(dagPath);	
            }

        }
        list = list1;
    }
    else 
    {
        for ( ; !dagIter.isDone(); dagIter.next() )
        {
            if (dagIter.getPath (dagPath) != MS::kSuccess) 
                continue;

            Obj = dagPath.node();
            MFnAttribute Attr(Obj);
            //if(Attr.isHidden()) {MessageBox(NULL,"1","2",MB_OK); continue;}

            list.add (dagPath);
        }
    }



    for(unsigned int i = 0;i<list.length();i++)
    {	
        list.getDagPath(i,dagPath);

        MFnDagNode dagNode( dagPath, &stat );

        if ( dagNode.isIntermediateObject() )
          continue;
        
        // We want only the shape, not the transform-extended-to-shape (OBJ)
        if ( dagPath.hasFn(MFn::kMesh) && dagPath.hasFn(MFn::kTransform) )
          continue;
        
        if ( !dagPath.hasFn( MFn::kMesh ) )
          continue;

        if ( !createDrawables(scene_geode, dagPath) )
            return  false;
    }	

    return true;
}

//---------------------------------------------------------------------------
bool OSGTranslator::createDrawables(
                                    osg::ref_ptr<osg::Geode> &geode,
                                    MDagPath dagPath
                                    )
{
    using namespace std;
    using namespace osg;
    
    cout << "createDrawables.." << endl;
    
    try
    {
        MStatus status;
        MDagPath path = dagPath;

        int instanceNum = 0;
        if (path.isInstanced())
            instanceNum = path.instanceNumber();

        MFnMesh fnMesh(path);
        string mesh_name = fnMesh.name().asChar();

        log_file << "*** Processing mesh node \"" << mesh_name
                 << "\" ***" << endl;

        MObjectArray sets;
        MObjectArray comps;
        if ( !fnMesh.getConnectedSetsAndMembers(instanceNum, sets, comps, true))
            log_file << "ERROR: MFnMesh::getConnectedSetsAndMembers() "
                     << instanceNum << endl;

       /*
       ADDITION_UV_SETS addition_uv_set = NONE;

        MStringArray setNames;
        fnMesh.getUVSetNames(setNames);
        for (unsigned int idx = 0; idx < setNames.length(); ++idx)
        {
            if ( setNames[idx] == LIGHTMAP_UVSET_NAME )
            {
                addition_uv_set = LIGHT_MAP;
                log_file << "Has LightMap UVSet!" << endl;
                break;
            }
            else
                if ( setNames[idx] == AUX_DATA_ME_UVSET_NAME )
                {
                    addition_uv_set = AUX_DATA_ME;
                    log_file << "Has AuxDataME UVSet!" << endl;
                    break;
                }
        }
        */

        for ( unsigned i = 0; i < sets.length(); i++ )
        {
            log_file << "\n\tset " << i << endl;

            MObject set		= sets[i];
            MObject comp	= comps[i];

            MFnSet fnSet( set, &status );
            if ( status == MS::kFailure )
            {
                log_file << "ERROR: MFnSet::MFnSet\n";
                continue;
            }

            // Make sure the set is a polygonal set.  If not, continue.
            MItMeshPolygon piter(path, comp, &status);
            if ( (status == MS::kFailure))
            {
                if (status == MS::kFailure)
                    log_file << "WARNING: This set is NOT polygonal!"
                             << "/ piter: status == MS::kFailure)" << endl;

                if (comp.isNull())
                    log_file << "piter: comp.isNull()" << endl;

                continue;
            }


            int num_polygons = piter.count();
            log_file << "Num polygons in this set: = " << num_polygons << endl;

            bool has_texture = true;

            MObject shaderNode = findShader(set);
            if (shaderNode == MObject::kNullObj)
            {
                log_file << "WARNING: This set doesn't have shader node!"
                         << "/ shaderNode == MObject::kNullObj" << endl;
                continue;
            }

            MFnDependencyNode fnShaderDN(shaderNode);
            log_file << "ShaderNodeName: " << fnShaderDN.name().asChar()
                     << endl;

#if 0
            if (fnShaderDN.name() == "lambert1")
            {
                log_file << "\"lambert1\" is a Maya default shader, it was skipped becouse of potential problems" << endl;
//                continue;
            }
#endif

            MPlug colorPlug = MFnDependencyNode(shaderNode).findPlug("color", &status);
            if (status == MS::kFailure)
            {
                log_file << "WARNING: Shader node of this set doesn't have \"color\" plug! / colorPlug: status == MS::kFailure" << endl;
                continue;
            }

            double r, g, b;
            colorPlug.child(0).getValue( r );
            colorPlug.child(1).getValue( g );
            colorPlug.child(2).getValue( b );

            color = Vec3(r,g,b);

            MItDependencyGraph dgIt(colorPlug, MFn::kFileTexture,
                MItDependencyGraph::kUpstream, 
                MItDependencyGraph::kBreadthFirst,
                MItDependencyGraph::kNodeLevel, 
                &status);

            if (status == MS::kFailure)
            {
                log_file << "WARNING: dgIt(colorPlug, MFn::kFileTexture,...):"
                         << " status == MS::kFailure" << endl;
                continue;
            }

            dgIt.disablePruningOnFilter();

            // If no texture file node was found, just continue.
            //
            if (dgIt.isDone())
            {
                log_file << "WARNING: No texture file node was found!" << endl;
                has_texture = false;
            }

            // Print out the texture node name and texture file
            // that it references.
            MString textureName;
            if (has_texture)
            {
                MObject textureNode = dgIt.thisNode();
                MPlug filenamePlug =
                  MFnDependencyNode(textureNode).findPlug("fileTextureName",
                                                          &status);
                if ( status != MS::kSuccess )
                    log_file << "WARNING: Plug \"fileTextureName\" not found!"
                             << endl;
                else
                    filenamePlug.getValue(textureName);

             }

            string print_tex_name = textureName.asChar();
            if ( print_tex_name.empty() )
                print_tex_name = "none";

            log_file << "Texture file name: " << print_tex_name << endl;

            //-----------------------------------------------
            //create data arrays for cur drawable
            //-----------------------------------------------

            log_file << "0 ";
            std::set<osg::Vec3>	drw_vertices;
            std::set<osg::Vec3>	drw_normals;
            std::set<osg::Vec2>	drw_tex_coords;
            log_file << "1 ";
            std::set<VertexAttributes> vertices;
            vector<VertexAttributes> triangle(3);
            vector<vector<VertexAttributes> > triangles;
            log_file << "2 ";
            Vec3	vec3;
            Vec2	vec2;
            MPoint	point;
            MVector	vector;
            float u, v;
            log_file << "3 ";

            for ( ; !piter.isDone(); piter.next() )
            {
                MPointArray face_points;
                piter.getPoints(face_points, MSpace::kWorld);

                if (face_points.length() != 3)
                {
                    log_file << "WARNING: face points count = "
                             << face_points.length() << endl;

                    is_scene_correct = false;
                    string node_info;
                    node_info += "Node \"";
                    node_info += fnMesh.name().asChar();
                    node_info += "\" has ";
                    node_info += int2string(num_polygons);
                    node_info += " polygons with points count ";
                    node_info += int2string(face_points.length());
                    node_info += ", mapped on texture \"";
                    node_info += textureName.asChar();
                    node_info += "\"";

                    non_triangulated_nodes.push_back( node_info );
                    break;
                }

                for (int i = 0; i < 3; ++i)
                {
                    assign(vec3, face_points[i]);
                    triangle[i].Coord = vec3;
                }

                MVectorArray face_normals;
                piter.getNormals(face_normals, MSpace::kWorld);
                for (int i = 0; i < 3; ++i)
                {
                    assign(vec3, face_normals[i]);
                    triangle[i].Normal = vec3;
                }

                MFloatArray face_Us, face_Vs;
                piter.getUVs(face_Us, face_Vs);
                if (face_Us.length() != 3 || face_Vs.length() != 3)
                {
                    triangle[0].TexCoord = Vec2();
                    triangle[1].TexCoord = Vec2();
                    triangle[2].TexCoord = Vec2();
                    log_file << "WARNING: found unmapped polygon N"
                             << piter.index() << endl;
                }
                else
                {
                    for (int i = 0; i < 3; ++i)
                    {
                        u = face_Us[i];
                        v = face_Vs[i];
                        vec2[0] = u;
                        vec2[1] = v;

                        triangle[i].TexCoord = vec2;
                    }
                }

                /*
                if ( addition_uv_set != NONE )
                {
                    if (addition_uv_set == LIGHT_MAP)
                        piter.getUVs(face_Us, face_Vs, &LIGHTMAP_UVSET_NAME);
                    else
                        if (addition_uv_set == AUX_DATA_ME)
                            piter.getUVs(face_Us, face_Vs, &AUX_DATA_ME_UVSET_NAME);

                    if (face_Us.length() != 3 || face_Vs.length() != 3)
                    {
                        triangle[0].TexCoord = Vec2();
                        triangle[1].TexCoord = Vec2();
                        triangle[2].TexCoord = Vec2();
                        log_file << "WARNING: found unmapped addition UV set polygon N" << piter.index() << endl;
                    }
                    else
                    {
                        for (int i = 0; i < 3; ++i)
                        {
                            u = face_Us[i];
                            v = face_Vs[i];
                            vec2[0] = u;
                            vec2[1] = v;

                            triangle[i].TexCoord1 = vec2;
                        }
                    }
                }	
                */

                vertices.insert( triangle[0] );
                vertices.insert( triangle[1] );
                vertices.insert( triangle[2] );
                triangles.push_back(triangle);
            }

            Material material;
            material.TextureName = textureName.asChar();

            log_file << "before createosggeometry_dupl.. " << endl;
            osg::ref_ptr<osg::Geometry> geom =
                createOSGGeometry_DuplicateCoords(fnMesh, /*addition_uv_set,*/
                                                 vertices, triangles, material);
           
            geode->addDrawable( geom.get() );
            
            log_file << "after createosggeometry_dupl.. " << endl;
        }

        log_file << "\n*** Mesh node \"" << mesh_name << "\" processed ***\n\n";
    }
    catch (...)
    {
        log_file << " ERROR: UNKNOWN ERROR!\n\n";
    }
    return true;
}

//---------------------------------------------------------------------------
osg::ref_ptr<osg::Geometry>
OSGTranslator::createOSGGeometry_DuplicateCoords(
    MFnMesh						&fnMesh,
    //ADDITION_UV_SETS			addition_uv_set,
    std::set<VertexAttributes>	&vertices,
    std::vector<std::vector<VertexAttributes> > &triangles,
    Material &material)
{
    log_file << "OSGGeometry creating started (duplicate coords mode)."
             << endl;
    MStatus status;

    using namespace osg;
    osg::ref_ptr<osg::Geometry> geom = new osg::Geometry();

    geom->setUseDisplayList(false);
    geom->setUseVertexBufferObjects(true);
    geom->setDataVariance(Object::STATIC);

    int ntris = triangles.size();
    bool is_u_short_index = (ntris * 3) < 65536;

    osg::ref_ptr<osg::DrawElementsUInt>		triangles_uint;
    osg::ref_ptr<osg::DrawElementsUShort>	triangles_ushort;

    if ( param_use_short_if_possible && is_u_short_index )
    {
        triangles_ushort =
          new osg::DrawElementsUShort(osg::PrimitiveSet::TRIANGLES);
        triangles_ushort->reserve(ntris * 3);
    }
    else
    {
        triangles_uint =
          new osg::DrawElementsUInt(osg::PrimitiveSet::TRIANGLES);
        triangles_uint->reserve(ntris * 3);
    }

    std::set<VertexAttributes>::iterator	vitr;

    map< VertexAttributes, int >			vertices_ids_map;

    int unique_vertex_attribute_count = vertices.size();
    
    osg::ref_ptr<osg::Vec3Array> drw_vertices_array =
      new osg::Vec3Array(unique_vertex_attribute_count);
    
    osg::ref_ptr<osg::Vec3Array> drw_normals_array =      
      new osg::Vec3Array(unique_vertex_attribute_count);
    
    osg::ref_ptr<osg::Vec2Array> drw_tex_coords_array	=
      new osg::Vec2Array(unique_vertex_attribute_count);
    
    osg::ref_ptr<osg::Vec2Array> drw_tex_coords_array1 =
      new osg::Vec2Array(unique_vertex_attribute_count);

    int last_texcoord_array_idx = 0;
    int idx;
    unsigned int i, j;

    log_file << "0 " << endl;

    for (i = 0, vitr = vertices.begin(); vitr != vertices.end(); ++vitr, ++i)
    {
        vertices_ids_map[*vitr] = i;

        (*drw_vertices_array)  [i]	 = vitr->Coord;
        (*drw_normals_array)   [i]	 = vitr->Normal;
        (*drw_tex_coords_array)[i]	 = vitr->TexCoord;
        (*drw_tex_coords_array1)[i]	 = vitr->TexCoord1;
    }

    log_file << "1 " << endl;

    for (i = 0; i < triangles.size(); ++i)
    {
        for (j = 0; j < 3; ++j)
        {
            idx = vertices_ids_map[ triangles[i][j] ];

            if ( param_use_short_if_possible && is_u_short_index )
                triangles_ushort->push_back(idx);
            else
                triangles_uint->push_back(idx);
        }
    }

    log_file << "2 " << endl;

    geom->setVertexArray(drw_vertices_array.get());
    geom->setNormalArray(drw_normals_array.get());
    geom->setTexCoordArray(0, drw_tex_coords_array.get());

    geom->setNormalBinding(osg::Geometry::BIND_PER_VERTEX);

    if ( param_use_short_if_possible && is_u_short_index )
        geom->addPrimitiveSet(triangles_ushort.get());
    else
        geom->addPrimitiveSet(triangles_uint.get());

    // temporary color array for lighting
    ref_ptr<Vec4Array> color_array = new Vec4Array();
    //color_array->push_back(Vec4(1.0f, 1.0f, 1.0f, 1.0f));
    //
    // sorry no transparency for now :p
    log_file << "color : " << color[0] << ", " << color[1] << ", " << color[2]
             << endl;
    color_array->push_back(Vec4( color[0], color[1], color[2], 1.0f));
    geom->setColorBinding(Geometry::BIND_OVERALL);
    geom->setColorArray(color_array.get());

#   if 0
    string lightmapped_mesh_name;
    string lightmapped_mesh_transform_name;

    if (addition_uv_set == LIGHT_MAP)
    {
        MObject node_obj = fnMesh.object();
        if ( node_obj.isNull() )
            log_file << "node_obj = NULL ";

        MObject attr_obj = fnMesh.attribute("lmapFileName");
        if ( attr_obj.isNull() )
            log_file << "attr_obj = NULL ";

        MPlug lmapFileNamePlug(node_obj, attr_obj);
        MString value;
        lmapFileNamePlug.getValue(value);

        string lightmap_name;
        lightmap_name = value.asChar();
        
        //trick to avoid much changes in code
        lightmapped_mesh_name = lightmap_name; 

        if (!lightmapped_mesh_name.empty()) 
            geom->setTexCoordArray(3, drw_tex_coords_array1.get());
    }
#   endif

#   if 1
    log_file << "3 " << endl;

    ref_ptr<osg::StateSet> stateset = createStateSetWithTexture(
        material.TextureName,
        param_texture_repeat_mode );
    if ( stateset.get()	)
        geom->setStateSet(stateset.get());

    log_file << "4 " << endl;
#   endif

#   if 0
    //tangent space
    if ( param_generate_tangent_space )
    {
        int normalunit = 0; // general TC 
        osg::ref_ptr<osgUtil::TangentSpaceGenerator> tsg = new osgUtil::TangentSpaceGenerator;
        tsg->generate(geom.get(), normalunit); 

        geom->setTexCoordArray (1, ::GetVec3ArrayFromVec4Array(tsg->getTangentArray()));

        if ( param_use_binormal_vector )
        {
            geom->setTexCoordArray (2, tsg->getBinormalArray()); 
        }
    }
#   endif

    log_file << "5 ";

    log_file << "geometry created." << endl;	
    return geom;
}


//---------------------------------------------------------------------------
osg::ref_ptr<osg::StateSet>
OSGTranslator::createStateSetWithTexture( const std::string &load_name,
                                          bool texture_repeat )
{
  log_file << "createStateSetWithTexture" << endl;
	using namespace osg;
	ref_ptr<osg::StateSet> stateset = new osg::StateSet;

  if( !load_name.empty() ) {
    osg::ref_ptr<osg::Texture2D> texture =
      createTexture(load_name, texture_repeat);

    if (texture.valid())
    {
      stateset->setTextureAttributeAndModes(0,texture.get(),StateAttribute::ON);

      // as in maya, texture replaces base color..
      osg::TexEnv* texenv = new osg::TexEnv;
      texenv->setMode(osg::TexEnv::REPLACE);
     
      stateset->setTextureAttribute(0,texenv);
    }
  }

  /*
	log_file << "T2 ";
	if ( lightmaped_mesh_name.empty() )
		return stateset;

	string replacing_name_for_litemap = GetCorrectedLightmapName(lightmaped_mesh_name);  //actually, it is a lightmap name
	log_file << replacing_name_for_litemap << " " << endl;
	log_file << "T3 ";
	osg::ref_ptr<osg::Texture2D> texture_litemap = GetLoadedTexture("", param_texture_repeat_mode, replacing_name_for_litemap);

    log_file << "T4 ";
	if ( texture_litemap.valid() )
	{
		stateset->setTextureAttributeAndModes(3, texture_litemap.get(), StateAttribute::ON);
	}
	log_file << "T5 ";
  */

  log_file << "Done.." << endl;
	return stateset;
}
//---------------------------------------------------------------------------
osg::ref_ptr<osg::Texture2D>
OSGTranslator::createTexture(const std::string& load_name, bool repeat_texture )
{
  log_file << "createTexture" << endl;
	using namespace osg;
	ref_ptr<Texture2D> texture;

  // file name without full path
  string name;
  
  // win32 file name
  if( name.find(":") != string::npos && name.find("\\") ) {
    name = strrchr(load_name.c_str(), '\\') + 1; // just the file name
  }
  else {
    name = strrchr(load_name.c_str(), '/') + 1; // just the file name
  }
  
  log_file << "Begin texture file " << load_name << " processing : "
           << name << endl;

	// load image
	Image *image = new Image;
  image->setFileName( name );

	/*image->setFileName(e_name);
	string texture_extension = short_file_name.substr(short_file_name.length() - 3, 3);*/

	// create texture
	texture = new osg::Texture2D;
	texture->setImage(image);

	if (param_texture_repeat_mode)
	{
		texture->setWrap(osg::Texture::WRAP_S, osg::Texture::REPEAT);
		texture->setWrap(osg::Texture::WRAP_T, osg::Texture::REPEAT);
		texture->setWrap(osg::Texture::WRAP_R, osg::Texture::REPEAT);
	}

    texture->setUnRefImageDataAfterApply(true);
	return texture;
}


#if 0
//---------------------------------------------------------------------------
MObject OSGTranslator::createCube()
{
    int						numVertices;
    float					cubeSize;
    MFloatPointArray		points;
    MFnMesh					meshFS;
    MObject					newMesh;
    // Scale the cube on the frame number, wrap every 10 frames.
    cubeSize		    		= 2.0f;

    const int numFaces			= 6;
    numVertices					= 8;
    const int numFaceConnects	= 24;

    MFloatPoint vtx_1( -cubeSize, -cubeSize, -cubeSize );
    MFloatPoint vtx_2(  cubeSize, -cubeSize, -cubeSize );
    MFloatPoint vtx_3(  cubeSize, -cubeSize,  cubeSize );
    MFloatPoint vtx_4( -cubeSize, -cubeSize,  cubeSize );
    MFloatPoint vtx_5( -cubeSize,  cubeSize, -cubeSize );
    MFloatPoint vtx_6( -cubeSize,  cubeSize,  cubeSize );
    MFloatPoint vtx_7(  cubeSize,  cubeSize,  cubeSize );
    MFloatPoint vtx_8(  cubeSize,  cubeSize, -cubeSize );

    points.append( vtx_1 );
    points.append( vtx_2 );
    points.append( vtx_3 );
    points.append( vtx_4 );
    points.append( vtx_5 );
    points.append( vtx_6 );
    points.append( vtx_7 );
    points.append( vtx_8 );

    // Set up an array containing the number of vertices
    // for each of the 6 cube faces (4 verticies per face)
    //
    int face_counts[numFaces] = { 4, 4, 4, 4, 4, 4 };
    MIntArray faceCounts( face_counts, numFaces );

    // Set up and array to assign vertices from points to each face 
    //
    int face_connects[ numFaceConnects ] = {	0, 1, 2, 3,
        4, 5, 6, 7,
        3, 2, 6, 5,
        0, 3, 5, 4,
        0, 4, 7, 1,
        1, 7, 6, 2	};
    MIntArray faceConnects( face_connects, numFaceConnects );

    newMesh = meshFS.create(numVertices, numFaces,
        points, faceCounts, faceConnects);

    return newMesh;
}
//---------------------------------------------------------------------------
MObject OSGTranslator::createPyramid()
{
    int						numVertices;
    float					cubeSize;
    MFloatPointArray		points;
    MFnMesh					meshFS;

    // Scale the cube on the frame number, wrap every 10 frames.
    cubeSize		    		= 2.0f;

    const int numFaces			= 4;
    numVertices					= 4;
    const int numFaceConnects	= 12;

    MFloatPoint vtx_1( 0.0f, 0.0f, 0.0f );
    MFloatPoint vtx_2( 0.0f, 0.0f, cubeSize );
    MFloatPoint vtx_3( 0.0f, cubeSize, 0.0f );
    MFloatPoint vtx_4( cubeSize, 0.0f, 0.0f );

    points.append( vtx_1 );
    points.append( vtx_2 );
    points.append( vtx_3 );
    points.append( vtx_4 );


    // Set up an array containing the number of vertices
    // for each of the 6 cube faces (4 verticies per face)

    MIntArray faceCounts( 4, 3 );

    int face_connects[ numFaceConnects ] = {	2, 3, 1,
        2, 1, 0,
        2, 0, 3,
        3, 0, 1	};


    MIntArray faceConnects;
    faceConnects.setLength( 12 );
    for (int i = 0; i < 12; ++i)
        faceConnects[i] = face_connects[i];

    MObject newMesh = meshFS.create(numVertices, numFaces,
        points, faceCounts, faceConnects);
    //									outData, &stat);

    return newMesh;
}


//---------------------------------------------------------------------------
void OSGTranslator::OSGRead(const char *fname)
{
    using namespace std;
    using namespace osg;

    log_file << "reading file: " << fname << endl;

    int  nameLength = strlen(fname);
    bool is_3df = (nameLength > 4) && !strcasecmp(fname + nameLength - 4, ".3df");
    bool is_osg = (nameLength > 4) && !strcasecmp(fname + nameLength - 4, ".osg");

    ref_ptr<osg::Node> node;

    if ( is_osg )
    {
        node = osgDB::readNodeFile( fname );
    }
    else
        if ( is_3df )
        {
            ref_ptr<osg::Geode> geode = Read_3df( fname, param_texture_repeat_mode ); 
            node = dynamic_cast<Node*>( geode.get() );
        }

        vector<Drawable*> drawables;
        ::GetDrawablesFromNode(node.get(), drawables);

        for (unsigned int i = 0; i < drawables.size(); ++i)
        {
            log_file << "drawable N" << i << endl;
            createMeshFromDrawable( drawables[i] );
        }
        log_file << "finish importing" << endl;
}

#endif

#if 0
//---------------------------------------------------------------------------
MObject OSGTranslator::createMeshFromDrawable(osg::Drawable *drw)
{
    using namespace std;
    using namespace osg;

    MObject newMesh;
    MStatus status;

    Geometry *geom = drw->asGeometry();
    if (!geom)
        return newMesh;

    int							i, j, k;
    MFloatPointArray			points;
    MFloatPoint					point;
    MFloatArray					u_array, v_array;
    MFnMesh						fnMesh;

    map<Vec3, int>				vtx_idx_map;
    map<Vec3, int>::iterator	mvec3itr;

    map<Vec2, int>				uv_idx_map;
    map<Vec2, int>::iterator	mvec2itr;

    const PrimitiveSet *primitive_set	= geom->getPrimitiveSet(0);
    const int num_triangles				= primitive_set->getNumPrimitives();
    const int num_vertex_indices		= primitive_set->getNumIndices();

    log_file << "num_triangles = " << num_triangles << endl;
    log_file << "num_vertex_indices = " << num_vertex_indices << endl;
    //-----------------------

    //number of vertices per each face. So we assume that OSG model contains only triangles
    MIntArray face_counts(num_triangles, 3);

    MIntArray face_connects;
    face_connects.setLength( num_vertex_indices);

    //-----------------------

    const Vec3Array* vertex_array	 = dynamic_cast<const Vec3Array*>(geom->getVertexArray());
    const Vec3Array* normal_array	 = dynamic_cast<const Vec3Array*>(geom->getNormalArray());
    const Vec2Array* texcoord_array	 = dynamic_cast<const Vec2Array*>(geom->getTexCoordArray(0));
    const IndexArray* index_array	 = geom->getVertexIndices();

    int idx = 0;

    log_file << "drawable contains: " << endl;
    log_file << "vertex count: " << vertex_array->size() << endl;
    log_file << "texcoords count: " << texcoord_array->size() << endl;
    log_file << "normals count: " << normal_array->size() << endl;

    //create set of unique points coords
    for (i = 0; i < num_vertex_indices; ++i)
    {
        idx = primitive_set->index(i);
        vtx_idx_map[ (*vertex_array)[idx] ] = 0;
    }
    log_file << "1" << endl;


    //fill in points array by unque set of coords and numerate these coords
    points.setLength( vtx_idx_map.size() );

    for (mvec3itr = vtx_idx_map.begin(), i = 0; mvec3itr != vtx_idx_map.end(); ++mvec3itr, ++i)
    {
        mvec3itr->second = i;
        assign( point, mvec3itr->first );
        points[i] = point;
        log_file << "point = " << mvec3itr->first << ", idx = " << i << endl;
    }
    log_file << "2" << endl;


    // Set up an array to assign vertices from points to each face
    for (i = 0; i < num_vertex_indices; ++i)
    {
        idx = primitive_set->index(i);
        face_connects[i] = vtx_idx_map[ (*vertex_array)[idx] ];
    }
    log_file << "3" << endl;


    //processing UVs
    for (i = 0; i < (int)texcoord_array->size(); ++i)
        uv_idx_map[ (*texcoord_array)[i] ] = 0;

    log_file << "4" << endl;

    //fill in UVs array
    u_array.setLength( uv_idx_map.size() );
    v_array.setLength( uv_idx_map.size() );

    log_file << "tex coords count = " << uv_idx_map.size() << endl;

    for (mvec2itr = uv_idx_map.begin(), i = 0; mvec2itr != uv_idx_map.end(); ++mvec2itr, ++i)
    {
        mvec2itr->second = i;
        u_array[i] = mvec2itr->first[0];
        v_array[i] = mvec2itr->first[1];
        log_file << "UV = " << u_array[i] << ", " << v_array[i] << endl;
    }

    log_file << "5" << endl;

    newMesh = fnMesh.create(points.length(), num_triangles,
        points, face_counts, face_connects);

    MString newMesh_name = fnMesh.setName("meshOSG#");
    log_file << "mesh created with name " << newMesh_name.asChar() << endl;


    status = fnMesh.setUVs(u_array, v_array);
    if (status != MS::kSuccess )
        log_file << "setUVs failed" << endl;

    MIntArray uv_counts(num_triangles, 3);
    MIntArray uv_ids;
    uv_ids.setLength(num_vertex_indices);

    int osg_idx;
    Vec2 vec2;

    for (i = 0; i < num_vertex_indices; ++i)
    {
        osg_idx = primitive_set->index(i);
        vec2 = (*texcoord_array)[osg_idx];
        uv_ids[i] = uv_idx_map[ vec2 ];
    }

    status = fnMesh.assignUVs(uv_counts, uv_ids);
    if (status != MS::kSuccess )
        log_file << "assignUVs failed" << endl;

    log_file << "UVs added" << endl;


    MVectorArray	normals;
    MIntArray		face_list;
    MIntArray		vertex_list;

    MVector			normal;

    int count = num_triangles * 3;
    normals.setLength(count);
    face_list.setLength(count);
    vertex_list.setLength(count);

    for (i = 0; i < num_triangles; ++i)
    {
        for (j = 0; j < 3; ++j)
        {
            k = i*3 + j;
            idx = primitive_set->index(k);

            face_list	[k] = i;
            vertex_list	[k] = face_connects[k];
            assign(normal, (*normal_array)[idx]);
            normals		[k] = normal;
        }
    }

    status = fnMesh.setFaceVertexNormals ( normals, face_list, vertex_list, MSpace::kObject);
    if (status != MS::kSuccess )
        log_file << "setFaceVertexNormals failed" << endl;

    log_file << "normals added" << endl;


    //processing texture
    StateSet* state_set = drw->getStateSet();
    if ( !state_set )
    {
        log_file << "there is no stateset" << endl;
        return newMesh;
    }

    StateAttribute* sa = state_set->getTextureAttribute(0, StateAttribute::TEXTURE);
    if (!sa)
    {
        log_file << "there is no StateAttribute::TEXTURE" << endl;
        return newMesh;
    }

    Texture2D* tex2D = dynamic_cast<Texture2D*>(sa);
    if (!sa)
    {
        log_file << "there is no Texture2D" << endl;
        return newMesh;
    }

    Image *image = tex2D->getImage();
    if (!image)
    {
        log_file << "there is no Image" << endl;
        return newMesh;
    }

    MString texture_file_name = image->getFileName().c_str();
    if ( !texture_file_name.length() )
    {
        log_file << "texture file name is empty!" << endl;
        texture_file_name = "d:/000.tga";
    }

    createMayaDrawableAPI(newMesh_name, texture_file_name);
    log_file << "mesh configured" << endl;
    return newMesh;
}
//---------------------------------------------------------------------------
void OSGTranslator::createMayaDrawableMEL(MString mesh_name, MString texture_name)
{
    MGlobal::executeCommand(MString("createNode lambert -n \"lambert2\""));
    MGlobal::executeCommand(MString("setAttr \".miic\" -type \"float3\" 3.1415927 3.1415927 3.1415927"));
    MGlobal::executeCommand(MString("createNode shadingEngine -n \"lambert2SG\""));
    MGlobal::executeCommand(MString("setAttr \".ihi\" 0"));
    MGlobal::executeCommand(MString("setAttr \".ro\" yes"));
    MGlobal::executeCommand(MString("createNode materialInfo -n \"materialInfo1\""));
    MGlobal::executeCommand(MString("createNode file -n \"file1\""));
    MGlobal::executeCommand(MString("setAttr \".ftn\" -type \"string\" \"") + texture_name.asChar() + "\"");
    MGlobal::executeCommand(MString("select -ne :renderPartition"));
    MGlobal::executeCommand(MString("setAttr -s 3 \".st\""));
    MGlobal::executeCommand(MString("select -ne :renderGlobalsList1"));
    MGlobal::executeCommand(MString("select -ne :defaultShaderList1"));
    MGlobal::executeCommand(MString("setAttr -s 3 \".s\""));
    MGlobal::executeCommand(MString("select -ne :postProcessList1"));
    MGlobal::executeCommand(MString("setAttr -s 2 \".p\""));
    MGlobal::executeCommand(MString("select -ne :defaultRenderUtilityList1"));
    MGlobal::executeCommand(MString("select -ne :lightList1"));
    MGlobal::executeCommand(MString("select -ne :defaultTextureList1"));

    MGlobal::executeCommand(MString("connectAttr \":defaultLightSet.msg\" \"lightLinker1.lnk[2].llnk\""));
    MGlobal::executeCommand(MString("connectAttr \"lambert2SG.msg\" \"lightLinker1.lnk[2].olnk\""));
    MGlobal::executeCommand(MString("connectAttr \"file1.oc\" \"lambert2.c\""));

    MGlobal::executeCommand(MString("connectAttr \"lambert2.oc\" \"lambert2SG.ss\""));
    MGlobal::executeCommand(MString("connectAttr \"") + mesh_name.asChar() + ".iog\" \"lambert2SG.dsm\" -na");
    MGlobal::executeCommand(MString("connectAttr \"lambert2SG.msg\" \"materialInfo1.sg\""));
    MGlobal::executeCommand(MString("connectAttr \"lambert2.msg\" \"materialInfo1.m\""));
    MGlobal::executeCommand(MString("connectAttr \"file1.msg\" \"materialInfo1.t\" -na"));
    MGlobal::executeCommand(MString("connectAttr \"lambert2SG.pa\" \":renderPartition.st\" -na"));
    MGlobal::executeCommand(MString("connectAttr \"lambert2.msg\" \":defaultShaderList1.s\" -na"));

    MGlobal::executeCommand(MString("connectAttr \"file1.msg\" \":defaultTextureList1.tx\" -na"));
}
//---------------------------------------------------------------------------
void OSGTranslator::createMayaDrawableMEL1(MString mesh_name, MString texture_name)
{

}
//---------------------------------------------------------------------------
void OSGTranslator::createMayaDrawableAPI(MString mesh_name, MString texture_name)
{
    MString name_tmp;
    MStatus status;
    MFnLambertShader fnLambert;

    MObject lambert_shader_node = fnLambert.create();
    MString lambert_shader_name = fnLambert.setName("lambertOSG#");
 
    MFnDependencyNode fnShadingEngine;
    MString shading_group_name = lambert_shader_name + "SG#";
    MObject shading_group_node = fnShadingEngine.create("shadingEngine", &status);
    shading_group_name = fnShadingEngine.setName(shading_group_name);
    MPlug	plug_isHistoricallyInteresting(shading_group_node, fnShadingEngine.attribute( MString("isHistoricallyInteresting") ) );
    plug_isHistoricallyInteresting.setValue( 0 );
    MPlug	plug_renderableOnlySet(shading_group_node, fnShadingEngine.attribute( MString("renderableOnlySet") ) );
    plug_renderableOnlySet.setValue( true );
    //--------
    //--------
    MString material_info_name = lambert_shader_name + "MI";
    MGlobal::executeCommand(MString("createNode materialInfo -n \"") + material_info_name + "\"");
    //--------
    //--------
    MFnDependencyNode fnFile;
    MObject tex_file_node = fnFile.create("file", &status);
    MString tex_file_name = fnFile.setName("fileOSG#");
    MPlug tex_filename_plug( tex_file_node, fnFile.attribute( MString("fileTextureName") ) );
    tex_filename_plug.setValue( texture_name.asChar() );
    //--------
    //--------
    MGlobal::executeCommand(MString("select -ne :renderPartition"));
    MGlobal::executeCommand(MString("setAttr -s 3 \".st\""));
    MGlobal::executeCommand(MString("select -ne :renderGlobalsList1"));
    MGlobal::executeCommand(MString("select -ne :defaultShaderList1"));
    MGlobal::executeCommand(MString("setAttr -s 3 \".s\""));
    MGlobal::executeCommand(MString("select -ne :postProcessList1"));
    MGlobal::executeCommand(MString("setAttr -s 2 \".p\""));
    MGlobal::executeCommand(MString("select -ne :defaultRenderUtilityList1"));
    MGlobal::executeCommand(MString("select -ne :lightList1"));
    MGlobal::executeCommand(MString("select -ne :defaultTextureList1"));

    MGlobal::executeCommand(MString("connectAttr \"") + tex_file_name + ".oc\" \"" + lambert_shader_name + ".c\"");
    MGlobal::executeCommand(MString("connectAttr \"") + lambert_shader_name + ".oc\" \"" + shading_group_name + ".ss\"");
    MGlobal::executeCommand(MString("connectAttr \"") + mesh_name + ".iog\" \"" + shading_group_name + ".dsm\" -na");
    MGlobal::executeCommand(MString("connectAttr \"") + shading_group_name + ".msg\" \"" + material_info_name + ".sg\"");
    MGlobal::executeCommand(MString("connectAttr \"") + lambert_shader_name + ".msg\" \"" + material_info_name + ".m\"");
    MGlobal::executeCommand(MString("connectAttr \"") + tex_file_name + ".msg\" \"" + material_info_name + ".t\" -na");
    MGlobal::executeCommand(MString("connectAttr \"") + shading_group_name + ".pa\" \":renderPartition.st\" -na");
    
    MGlobal::executeCommand(MString("connectAttr \"") + tex_file_name + ".msg\" \":defaultTextureList1.tx\" -na");
}
#endif

#if 0
//---------------------------------------------------------------------------
void OSGTranslator::getMeshHoles(MFnMesh &fnMesh, vector<PolygonHoles> &holes)
{
    using namespace std;
    using namespace osg;

    holes.resize(fnMesh.numPolygons());
    vector<Vec3> empty_hole;

    MStatus status;

    MPointArray mesh_points;
    fnMesh.getPoints( mesh_points, MSpace::kWorld );
    Vec3 vec3;

    MIntArray holeInfoArray, holeVertexArray;
    fnMesh.getHoles ( holeInfoArray, holeVertexArray, &status );

    unsigned int holes_count = holeInfoArray.length() / 3;
    for (unsigned int i = 0; i < holes_count; ++i)
    {
        int polygon_idx			= holeInfoArray[i * 3];
        int hole_points_count	= holeInfoArray[i * 3 + 1];
        int start_idx			= holeInfoArray[i * 3 + 2];

        holes[polygon_idx].push_back(empty_hole);
        vector<Vec3> &hole_points = holes[polygon_idx].back();

        for (int j = 0; j < hole_points_count; ++j)
        {
            assign(vec3, mesh_points[holeVertexArray[start_idx + j]]);
            hole_points.push_back( vec3 );
        }
    }
}
//---------------------------------------------------------------------------
#endif


#if 0
//---------------------------------------------------------------------------
bool OSGTranslator::getMayaSceneAsOneDrawableOSGGeode(osg::ref_ptr<osg::Node> &scene_node)
{
    using namespace osg;
    using namespace std;

    osg::ref_ptr<osg::Geode> scene_geode = new osg::Geode();
    scene_node = scene_geode.get();
    scene_node->setName("Exported From Maya");

    non_triangulated_nodes.clear();
    is_scene_correct = true;

    vector<MDagPath> dag_paths;

    MStatus stat;
    MItDag dagIter( MItDag::kBreadthFirst, MFn::kInvalid, &stat );

    for ( ; !dagIter.isDone(); dagIter.next() )
    {
        MDagPath dagPath;
        stat = dagIter.getPath( dagPath );

        if ( stat )
        {
            MFnDagNode dagNode( dagPath, &stat );

            if ( dagNode.isIntermediateObject() ) continue;
            if ( dagPath.hasFn(MFn::kMesh) && dagPath.hasFn(MFn::kTransform) ) continue;
            if ( !dagPath.hasFn( MFn::kMesh ) ) continue;

            dag_paths.push_back(dagPath);
        }
    }


    GetAllInOneDrawable(scene_geode, dag_paths);

    return true;
}

//---------------------------------------------------------------------------
void OSGTranslator::CheckDagPath(MDagPath path)
{

}
//---------------------------------------------------------------------------
bool OSGTranslator::GetAllInOneDrawable(
                                        osg::ref_ptr<osg::Geode> &geode, std::vector<MDagPath> &dag_paths)
{
    using namespace std;
    using namespace osg;

    set<VertexAttributes>				vertices;
    vector<VertexAttributes>			triangle(3);
    vector<vector<VertexAttributes> >	triangles;
    string								texture_name;
    //ADDITION_UV_SETS					addition_uv_set = NONE;


    log_file << "creating AllInOneDrawable:" << endl;

    for (unsigned int i = 0; i < dag_paths.size(); ++i)
    {
        try
        {
            MStatus status;
            MDagPath path = dag_paths[i];

            int instanceNum = 0;
            if (path.isInstanced())
                instanceNum = path.instanceNumber();
            //
            MFnMesh fnMesh(path);
            string mesh_name = fnMesh.name().asChar();

            log_file << "*** Processing mesh node \"" << mesh_name << "\" ***" << endl;

            if (mesh_name.length() >= 4 && mesh_name.substr(0,4) == "tag_")
            {
                log_file << "Mesh node \"" << mesh_name << "\" is a tag so it was skipped!" << endl;
                continue;
            }

            MObjectArray sets;
            MObjectArray comps;
            if ( !fnMesh.getConnectedSetsAndMembers(instanceNum, sets, comps, true) )
                log_file << "ERROR: MFnMesh::getConnectedSetsAndMembers() " << instanceNum << endl;

            //--------
            //skipped light map processing
            //bool has_light_map_uvset = false;
            //--------

            for ( unsigned i = 0; i < sets.length(); i++ )
            {
                log_file << "\n\tset " << i << endl;

                MObject set		= sets[i];
                MObject comp	= comps[i];

                MFnSet fnSet( set, &status );
                if ( status == MS::kFailure )
                {
                    log_file << "ERROR: MFnSet::MFnSet\n";
                    continue;
                }

                // Make sure the set is a polygonal set.  If not, continue.
                MItMeshPolygon piter(path, comp, &status);
                if ( (status == MS::kFailure))
                {
                    if (status == MS::kFailure)
                        log_file << "WARNING: This set is NOT polygonal! / piter: status == MS::kFailure)" << endl;

                    if (comp.isNull())
                        log_file << "piter: comp.isNull()" << endl;

                    continue;
                }


                int num_polygons = piter.count();
                log_file << "Num polygons in this set: = " << num_polygons << endl;

                bool has_texture = true;

                MObject shaderNode = findShader(set);
                if (shaderNode == MObject::kNullObj)
                {
                    log_file << "WARNING: This set doesn't have shader node! / shaderNode == MObject::kNullObj" << endl;
                    continue;
                }

                MFnDependencyNode fnShaderDN(shaderNode);
                log_file << "ShaderNodeName: " << fnShaderDN.name().asChar() << endl;

                MPlug colorPlug = MFnDependencyNode(shaderNode).findPlug("color", &status);
                if (status == MS::kFailure)
                {
                    log_file << "WARNING: Shader node of this set doesn't have \"color\" plug! / colorPlug: status == MS::kFailure" << endl;
                    continue;
                }

                double r, g, b;
                plug.getChild(0).getValue( r );
                plug.getChild(1).getValue( g );
                plug.getChild(2).getValue( b );

                color = Vec3(r,g,b);

                MItDependencyGraph dgIt(colorPlug, MFn::kFileTexture,
                    MItDependencyGraph::kUpstream, 
                    MItDependencyGraph::kBreadthFirst,
                    MItDependencyGraph::kNodeLevel, 
                    &status);

                if (status == MS::kFailure)
                {
                    log_file << "WARNING: dgIt(colorPlug, MFn::kFileTexture,...): status == MS::kFailure" << endl;
                    continue;
                }

                dgIt.disablePruningOnFilter();

                // If no texture file node was found, just continue.
                //
                if (dgIt.isDone())
                {
                    log_file << "WARNING: No texture file node was found!" << endl;
                    has_texture = false;
                }

                // Print out the texture node name and texture file that it references.
                //
                MString textureName;
                if (has_texture)
                {
                    MObject textureNode = dgIt.thisNode();
                    MPlug filenamePlug = MFnDependencyNode(textureNode).findPlug("fileTextureName", &status);
                    if ( status != MS::kSuccess )
                        log_file << "WARNING: Plug \"fileTextureName\" not found!" << endl;
                    else
                        filenamePlug.getValue(textureName);
                }

                texture_name = textureName.asChar();

                string print_tex_name = textureName.asChar();
                if ( print_tex_name.empty() )
                    print_tex_name = "none";

                log_file << "Texture file name: " << print_tex_name << endl;

                //-----------------------------------------------
                //add current drawable data to arrays for all in one drawable
                //-----------------------------------------------

                Vec3	vec3;
                Vec2	vec2;
                MPoint	point;
                MVector	vector;
                float u, v;

                for ( ; !piter.isDone(); piter.next() )
                {
                    MPointArray face_points;
                    piter.getPoints(face_points, MSpace::kWorld);

                    if (face_points.length() != 3)
                    {
                        log_file << "WARNING: face points count = " << face_points.length() << endl;

                        //--------

                        is_scene_correct = false;
                        string node_info;
                        node_info += "Node \"";
                        node_info += fnMesh.name().asChar();
                        node_info += "\" has ";
                        node_info += int2string(num_polygons);
                        node_info += " polygons with points count ";
                        node_info += int2string(face_points.length());
                        node_info += ", mapped on texture \"";
                        node_info += textureName.asChar();
                        node_info += "\"";

                        non_triangulated_nodes.push_back( node_info );
                        break;
                    }

                    for (int i = 0; i < 3; ++i)
                        assign(triangle[i].Coord, face_points[i]);

                    MVectorArray face_normals;
                    piter.getNormals(face_normals, MSpace::kWorld);
                    for (int i = 0; i < 3; ++i)
                        assign(triangle[i].Normal, face_normals[i]);

                    MFloatArray face_Us, face_Vs;
                    piter.getUVs(face_Us, face_Vs);
                    if (face_Us.length() != 3 || face_Vs.length() != 3)
                    {
                        triangle[0].TexCoord = Vec2();
                        triangle[1].TexCoord = Vec2();
                        triangle[2].TexCoord = Vec2();
                    }
                    else
                    {
                        for (int i = 0; i < 3; ++i)
                        {
                            u = face_Us[i];
                            v = face_Vs[i];
                            vec2[0] = u;
                            vec2[1] = v;

                            triangle[i].TexCoord = vec2;
                        }
                    }

                    vertices.insert( triangle[0] );
                    vertices.insert( triangle[1] );
                    vertices.insert( triangle[2] );
                    triangles.push_back(triangle);
                }
            }

            log_file << "\n*** Mesh node \"" << mesh_name << "\" processed ***\n\n";
        }
        catch (...)
        {
            log_file << " ERROR: UNKNOWN ERROR!\n\n";
        }
    }

    Material material;
    material.TextureName = texture_name;

    MFnMesh dummy_fnMesh;

    osg::ref_ptr<osg::Geometry> geom =
        CreateOSGGeometry_DuplicateCoords(dummy_fnMesh, /*addition_uv_set,*/
            vertices, triangles, material);

    geode->addDrawable( geom.get() );
    return true;
}
#endif

