// Code released under the OSGPL Licence.
// by DioSoft (http://www.diosoft.com)

/*
                OpenSceneGraph Public License, Version 0.0
                ==========================================
                
  Copyright (C) 2002 Robert Osfield.
  
  Everyone is permitted to copy and distribute verbatim copies
  of this licence document, but changing it is not allowed.

                       OPENSCENEGRAPH PUBLIC LICENCE
     TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION

  This library is free software; you can redistribute it and/or modify it
  under the terms of the OpenSceneGraph Public License (OSGPL) version 0.0
  or later.
  
  Notes: the OSGPL is based on the LGPL, with the 4 exceptions laid in in
  the wxWindows section below.  The LGPL in the final section of this
  license.


-------------------------------------------------------------------------------

                wxWindows Library Licence, Version 3
                ====================================

  Copyright (C) 1998 Julian Smart, Robert Roebling [, ...]

  Everyone is permitted to copy and distribute verbatim copies
  of this licence document, but changing it is not allowed.

                       WXWINDOWS LIBRARY LICENCE
     TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION

  This library is free software; you can redistribute it and/or modify it
  under the terms of the GNU Library General Public Licence as published by
  the Free Software Foundation; either version 2 of the Licence, or (at
  your option) any later version.

  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library
  General Public Licence for more details.

  You should have received a copy of the GNU Library General Public Licence
  along with this software, usually in a file named COPYING.LIB.  If not,
  write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA 02111-1307 USA.

  EXCEPTION NOTICE

  1. As a special exception, the copyright holders of this library give
  permission for additional uses of the text contained in this release of
  the library as licenced under the wxWindows Library Licence, applying
  either version 3 of the Licence, or (at your option) any later version of
  the Licence as published by the copyright holders of version 3 of the
  Licence document.

  2. The exception is that you may use, copy, link, modify and distribute
  under the user's own terms, binary object code versions of works based
  on the Library.

  3. If you copy code from files distributed under the terms of the GNU
  General Public Licence or the GNU Library General Public Licence into a
  copy of this library, as this licence permits, the exception does not
  apply to the code that you add in this way.  To avoid misleading anyone as
  to the status of such modified files, you must delete this exception
  notice from such code and/or adjust the licensing conditions notice
  accordingly.

  4. If you write modifications of your own for this library, it is your
  choice whether to permit this exception to apply to your modifications.
  If you do not wish that, you must delete the exception notice from such
  code and/or adjust the licensing conditions notice accordingly.
*/
  
//------------------------------------------------------------------------------

#include <maya/MObject.h>
#include <maya/MDagPath.h>
#include <maya/MPxFileTranslator.h>
#include <maya/MFnMesh.h>
#include <maya/MItMeshPolygon.h>

#include <osg/Vec3>
#include <osg/Vec2>
#include <osg/Geometry>
#include <osg/Texture2D>

#include <set>
#include <string>
#include <vector>

//---------------------------------------------------------------------------
struct Material
{
  std::string TextureName;
};

//enum ADDITION_UV_SETS { NONE, LIGHT_MAP, AUX_DATA_ME };

//---------------------------------------------------------------------------

/*
 * class PolygonHoles : public std::vector< std::vector<osg::Vec3> >
{
};
*/

//---------------------------------------------------------------------------
struct VertexAttributes
{
    osg::Vec3 Coord;
    osg::Vec3 Normal;
    osg::Vec2 TexCoord;
    osg::Vec2 TexCoord1;

    bool operator < (const VertexAttributes &other) const
    {
        if (Coord < other.Coord) return true;
        else if (other.Coord < Coord ) return false;
        else if (TexCoord < other.TexCoord) return true;
        else if (other.TexCoord < TexCoord ) return false;
        else if (TexCoord1 < other.TexCoord1) return true;
        else if (other.TexCoord1 < TexCoord1 ) return false;
        else return (Normal < other.Normal);
    }
};


//---------------------------------------------------------------------------

//***************************************************************************
//************************** class OSGTranslator ****************************
//***************************************************************************

class OSGTranslator : public MPxFileTranslator {
public:
    OSGTranslator () {};
    virtual         ~OSGTranslator () {};
    static void*    creator();

#if 0
    MStatus         reader ( const MFileObject& file,
                             const MString& optionsString,
                             FileAccessMode mode);
#endif
    MStatus         writer ( const MFileObject& file,
                             const MString& optionsString,
                             FileAccessMode mode );
    bool            haveReadMethod () const;
    bool            haveWriteMethod () const;
    MString         defaultExtension () const;
    MString			    filter () const;
    MFileKind       identifyFile ( const MFileObject& fileName,
                                   const char* buffer,
                                   short size) const;

private:

    bool			param_texture_repeat_mode;
    
    bool			param_generate_tangent_space;
    bool			param_use_binormal_vector;
    
    bool			param_duplicate_coords;
    bool			param_use_short_if_possible;

    osg::Vec3 color;

    bool			is_scene_correct;
    std::vector<std::string>	non_triangulated_nodes;

    void			OSGWrite(const char *fname,FileAccessMode mode);
    //void			OSGRead(const char *fname);


    bool			getMayaSceneAsOSGNode(osg::ref_ptr<osg::Node> &scene_node,
                                    FileAccessMode mode);
    
    /*
    bool			GetAllInOneDrawable(osg::ref_ptr<osg::Geode> &geode,
                                  std::vector<MDagPath> &dag_paths);
    bool			getMayaSceneAsOneDrawableOSGGeode(
                                    osg::ref_ptr<osg::Node> &scene_node);
   */

    /*
    MObject			createCube();
    MObject			createPyramid();
    MObject			createMeshFromDrawable(osg::Drawable *drw);
    */

    //methods for working with materials
    MObject findShader( MObject& setNode );

    bool createDrawables(
        osg::ref_ptr<osg::Geode> &geode,
        MDagPath dagPath);

    osg::ref_ptr<osg::Geometry> createOSGGeometry_SeparateIndices(
        MItMeshPolygon &piter,
        std::set<osg::Vec3>	&drw_vertices,
        std::set<osg::Vec3>	&drw_normals,
        std::set<osg::Vec2>	&drw_tex_coords,
        Material &material);

    osg::ref_ptr<osg::Geometry>	createOSGGeometry_DuplicateCoords(
        MFnMesh						&fnMesh,
        //ADDITION_UV_SETS			addition_uv_set,
        std::set<VertexAttributes>	&vertices,
        std::vector<std::vector<VertexAttributes> > &triangles,
        Material					&material);

    osg::ref_ptr<osg::StateSet>
      createStateSetWithTexture(
          const std::string &load_name, bool repeat_texture );

    osg::ref_ptr<osg::Texture2D>
      createTexture( const std::string& load_name, bool repeat_texture );


    void getOptionsValues(const MString	&options_string);

    void getPointsArrayFromGeometry(
        MFloatPointArray &points, const osg::Geometry &geom);

    /*
    void createMayaDrawableMEL(MString mesh_name, MString texture_name);
    void createMayaDrawableMEL1(MString mesh_name, MString texture_name);
    void createMayaDrawableAPI(MString mesh_name, MString texture_name);
    */
    
    void outNonTriangulatedNodesError();
    //void CheckDagPath(MDagPath path);

    //void getMeshHoles(MFnMesh &fnMesh, std::vector<PolygonHoles> &holes);
};
